from housaku.db import db_connection


def search(
    sqlite_url: str, query: str, limit: int = 20
) -> list[tuple[str, str, str, str]]:
    with db_connection(sqlite_url) as conn:
        cursor = conn.cursor()
        sql_query = """
            SELECT uri, title, type, substr(body, 0, 300)
            FROM documents
            WHERE ROWID IN (
                SELECT ROWID
                FROM documents_fts
                WHERE documents_fts MATCH ?
                ORDER BY bm25(documents_fts)
                LIMIT ?
            )
            """

        cursor.execute(sql_query, (query, limit))
        results = cursor.fetchall()
        return results



import sqlite3
from contextlib import contextmanager


def init_db(sqlite_url: str) -> None:
    conn = sqlite3.connect(sqlite_url)
    cursor = conn.cursor()

    # Creates the documents table.
    cursor.execute("""
    CREATE TABLE IF NOT EXISTS documents (
        uri TEXT UNIQUE NOT NULL,
        title TEXT,
        type TEXT NOT NULL,
        body TEXT NOT NULL
    );
    """)

    # Adds index on the uri column
    cursor.execute("CREATE INDEX IF NOT EXISTS idx_uri ON documents(uri)")

    # Creates virtual FTS5 table for full-text search
    cursor.execute("""
    CREATE VIRTUAL TABLE IF NOT EXISTS documents_fts USING fts5 (
        uri,
        title,
        type,
        body,
        content=documents,
        tokenize="porter unicode61"
    );
    """)

    # Settings
    cursor.execute("PRAGMA journal_mode = WAL;")
    cursor.execute("PRAGMA foreign_keys = ON;")
    cursor.execute("PRAGMA synchronous = NORMAL;")
    cursor.execute("PRAGMA busy_timeout = 5000;")
    cursor.execute("PRAGMA temp_store = MEMORY;")
    cursor.execute("PRAGMA cache_size = 2000;")

    cursor.execute("PRAGMA transaction_mode = IMMEDIATE;")

    conn.commit()
    conn.close()


def purge_db(sqlite_url: str) -> None:
    conn = sqlite3.connect(sqlite_url)
    cursor = conn.cursor()

    cursor.execute("DROP TABLE IF EXISTS documents;")
    cursor.execute("DROP TABLE IF EXISTS documents_fts;")
    cursor.execute("DROP INDEX IF EXISTS idx_uri;")

    cursor.execute("VACUUM;")

    conn.commit()
    conn.close()


def rebuilt_fts_table(sqlite_url: str) -> None:
    conn = sqlite3.connect(sqlite_url)
    cursor = conn.cursor()

    cursor.execute("INSERT INTO documents_fts(documents_fts) VALUES('rebuild');")

    conn.commit()
    conn.close()


@contextmanager
def db_connection(sqlite_url: str):
    conn = sqlite3.connect(sqlite_url)

    try:
        yield conn
        conn.commit()
    except:
        conn.rollback()
        raise
    finally:
        conn.close()



import shutil
from importlib.metadata import version
from typing import Type, Tuple
from pathlib import Path
import click
from pydantic import BaseModel, DirectoryPath, Field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)

app_name = "housaku"
app_dir = Path(click.get_app_dir(app_name=app_name))
config_file_path = app_dir / "config.toml"
template_file_path = Path(__file__).parent / "config_template.toml"


class FileSettings(BaseModel):
    include: list[DirectoryPath] = []
    exclude: list[str] = []


class FeedSettings(BaseModel):
    urls: list[str] = []


class Settings(BaseSettings):
    name: str = app_name
    description: str = (
        "A powerful personal search engine built on top of SQLite's FTS5."
    )
    version: str = version("housaku")

    sqlite_url: str = f"{app_dir / 'db.sqlite3'}"
    files: FileSettings = Field(default_factory=FileSettings)
    feeds: FeedSettings = Field(default_factory=FeedSettings)

    model_config = SettingsConfigDict(
        toml_file=config_file_path,
        extra="ignore",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        app_dir.mkdir(parents=True, exist_ok=True)

        if not config_file_path.exists():
            shutil.copy(template_file_path, config_file_path)

        return (
            init_settings,
            TomlConfigSettingsSource(settings_cls),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )


if __name__ == "__main__":
    settings = Settings()
    print(settings.model_dump())



import textwrap
from time import perf_counter
import urllib.parse
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Container
from textual.reactive import reactive
from textual.validation import ValidationResult, Validator
from textual.widgets import Button, Footer, Input, ListItem, ListView, Static
from housaku.db import init_db
from housaku.files import FILETYPES_SET
from housaku.settings import Settings
from housaku.search import search


class SearchInputValidator(Validator):
    def validate(self, value: str) -> ValidationResult:
        return (
            self.success()
            if value
            else self.failure("Search query cannot be an empty string.")
        )


class HousakuApp(App):
    TITLE = "housaku"
    CSS_PATH = "global.tcss"
    ENABLE_COMMAND_PALETTE = False

    BINDINGS = [
        Binding("ctrl+q", "quit", "Quit", priority=True),
    ]

    search_query: reactive[str] = reactive("")

    def __init__(self, settings: Settings):
        self.settings = settings
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Container(
            Static(
                f"{self.settings.name}",
                classes="about__title",
            ),
            Static(
                f"{self.settings.version}",
                classes="about__version",
            ),
            classes="about",
        )
        yield Horizontal(
            Input(
                placeholder="Search",
                type="text",
                classes="search__input",
                validate_on=["submitted"],
                validators=[SearchInputValidator()],
            ),
            Button(
                label="Search",
                classes="search__button",
                disabled=True,
            ),
            classes="search",
        )
        yield ListView(classes="results")
        yield Footer(classes="footer")

    def on_mount(self) -> None:
        self.input_field = self.query_one(Input)
        self.search_button = self.query_one(Button)

        self.results = self.query_one(ListView)
        self.results.border_title = "results"

    @on(Input.Submitted)
    @on(Button.Pressed, selector=".search__button")
    def handle_submit(self, _: Input.Submitted | Button.Pressed) -> None:
        if not self.input_field.is_valid:
            self.notify(
                "Search query cannot be empty.",
                severity="error",
            )
            return

        self.search_documents(self.search_query)

    @on(Input.Changed)
    def handle_change_input(self, event: Input.Changed) -> None:
        self.search_query = event.value

    def watch_search_query(self, value: str) -> None:
        if value:
            self.search_button.disabled = False
        else:
            self.search_button.disabled = True

    def search_documents(self, query: str) -> None:
        self.results.loading = True
        self.results.clear()

        start_time = perf_counter()

        try:
            results = search(self.settings.sqlite_url, query, 10)
        except Exception as e:
            self.notify(
                f"Something went wrong with your query: {e}.",
                severity="error",
            )
            self.results.loading = False
            return

        end_time = perf_counter()

        if not results:
            self.notify(
                "No results found.",
                severity="warning",
            )
            self.results.loading = False
            return

        for uri, title, doc_type, content in results:
            encoded_uri = urllib.parse.quote(uri, safe=":/")
            doc_title = title if title else uri
            truncated_content = textwrap.shorten(content, width=280, placeholder="...")

            if doc_type in FILETYPES_SET:
                link = f"[link=file://{encoded_uri}]{uri}[/]"
            else:
                link = f"[link={uri}]{uri}[/]"

            self.results.mount(
                ListItem(
                    Static(doc_title, classes="result__title"),
                    Static(link, classes="result__link"),
                    Static(truncated_content, classes="result__content"),
                    classes="result",
                )
            )

        elapsed_time = end_time - start_time
        self.results.loading = False
        self.results.focus()
        self.results.index = 0
        self.results.border_subtitle = f"in {elapsed_time:.3f}s"


if __name__ == "__main__":
    settings = Settings()
    init_db(settings.sqlite_url)

    app = HousakuApp(settings)
    app.run()
